#############################################################################
# Software Testing Automation Framework (STAF)                              #
# (C) Copyright IBM Corp. 2001                                              #
#                                                                           #
# This software is licensed under the Common Public License (CPL) V1.0.     #
#############################################################################

###################################################
# STAFUtil.tcl - STAF Utility TCL Library         #
###################################################

package provide STAF 3.0
package require TCLSTAF

namespace eval STAF {

    variable kOk                                 0
    variable kInvalidAPI                         1
    variable kUnknownService                     2
    variable kInvalidHandle                      3
    variable kHandleAlreadyExists                4
    variable kHandleDoesNotExist                 5
    variable kUnknownError                       6
    variable kInvalidRequestString               7
    variable kInvalidServiceResult               8
    variable kREXXError                          9
    variable kBaseOSError                       10
    variable kProcessAlreadyComplete            11
    variable kProcessNotComplete                12
    variable kVariableDoesNotExist              13
    variable kUnResolvableString                14
    variable kInvalidResolveString              15
    variable kNoPathToMachine                   16
    variable kFileOpenError                     17
    variable kFileReadError                     18
    variable kFileWriteError                    19
    variable kFileDeleteError                   20
    variable kSTAFNotRunning                    21
    variable kCommunicationError                22
    variable kTrusteeDoesNotExist               23
    variable kInvalidTrustLevel                 24
    variable kAccessDenied                      25
    variable kSTAFRegistrationError             26
    variable kServiceConfigurationError         27
    variable kQueueFull                         28
    variable kNoQueueElement                    29
    variable kNotifieeDoesNotExist              30
    variable kInvalidAPILevel                   31
    variable kServiceNotUnregisterable          32
    variable kServiceNotAvailable               33
    variable kSemaphoreDoesNotExist             34
    variable kNotSemaphoreOwner                 35
    variable kSemaphoreHasPendingRequests       36
    variable kTimeout                           37
    variable kJavaError                         38
    variable kConverterError                    39
    variable kInvalidObject                     41
    variable kInvalidParm                       42
    variable kRequestNumberNotFound             43
    variable kInvalidAsynchOption               44
    variable kRequestNotComplete                45
    variable kProcessAuthenticationDenied       46
    variable kInvalidValue                      47
    variable kDoesNotExist                      48
    variable kAlreadyExists                     49
    variable kDirectoryNotEmpty                 50
    variable kDirectoryCopyError                51
    variable kDiagnosticsNotEnabled             52
    variable kHandleAuthenticationDenied        53
    variable kHandleAlreadyAuthenticated        54

}

################################################################
# STAF::WrapData - Wraps a TCL String using colon delimited    # 
#                  STAF format                                 # 
# Accepts: A string                                            #
# Returns: Colon-delimited version of input string             #
################################################################

proc STAF::WrapData {data} { 
  return :[string length $data]:$data
}


################################################################################
# STAF::internalCheckCommandInput - Used internally by STAF to ensure a caller #
#                                   has provide proper parameters to a proc    #
# optionDefs:                                                                  #
#   {                                                                          #
#     { optionName ?default value? } ...                                       #
#   }                                                                          #
#                                                                              #
# Note: If no default value is provided then the option is not allowed to      #
#       have a value                                                           #
#                                                                              #
# argDefs:                                                                     #
#   {                                                                          #
#     { argName ?default value? } ...                                          #
#     ?otherArgs?                                                              #
#   }                                                                          #
#                                                                              #
# Note: If no default value is provided then the argument must be present      #
# Note: If an argName of "otherArgs" is specified, then all remaining args     #
#       will be grouped with that argument                                     #
#                                                                              #
# Returns:                                                                     #
#   {                                                                          #
#     { returnCode ?error string if non-zero RC? }                             #
#     ?{ optionName optionValue ... }?  -- if RC == 0                          #
#     ?{ argName argValue ... }?        -- if RC == 0                          #
#   }                                                                          #
#                                                                              #
# Note: Only options actually specified by the user are returned               #
################################################################################

proc STAF::internalCheckCommandInput {commandInput optionDefs argDefs} {
    # Initialize error buffers

    set errorBuffer {wrong # args: should be "}
    set validOptions {valid options are}

    # Set up option definitions

    foreach optionDef $optionDefs {
        if {[llength $optionDef] == 2} {
            set options([lindex $optionDef 0]) "DEFAULT [lindex $optionDef 1]"
            set errorBuffer "$errorBuffer ?-[lindex $optionDef 0] ?value??"
            set validOptions "$validOptions -[lindex $optionDef 0]"
        } else {
            set options([lindex $optionDef 0]) NOVALUEALLOWED
            set errorBuffer "$errorBuffer ?-[lindex $optionDef 0]?"
            set validOptions "$validOptions -[lindex $optionDef 0]"
        }
    }

    # Set up argument definitions

    set numArgsDefined 0

    foreach argDef $argDefs {
        if {[llength $argDef] == 2} {
            set args($numArgsDefined) [list [lindex $argDef 0] \
                                            "OPTIONAL [lindex $argDef 1]"]
            set errorBuffer "$errorBuffer ?[lindex $argDef 0]?"
        } else {
            set args($numArgsDefined) [list [lindex $argDef 0] REQUIRED]
            set errorBuffer "$errorBuffer [lindex $argDef 0]"
        }

        incr numArgsDefined
    }

    # Get all options out of the input

    while {[string index [lindex $commandInput 0] 0] eq "-"} {
        set currOption [string range [lindex $commandInput 0] 1 end]
        set currOptionValue {}
        set commandInput [lrange $commandInput 1 end]

        if {[info exists options($currOption)] == 0} {
            return [list [list 1 "bad option \"-$currOption\": $validOptions"]]
        } else {
            if {$options($currOption) eq "NOVALUEALLOWED"} {
                set outputOptions($currOption) {}
            } else {
                if {[string index [lindex $commandInput 0] 0] ne "-"} {
                    set outputOptions($currOption) [lindex $commandInput 0]
                    set commandInput [lrange $commandInput 1 end]
                } else {
                    set outputOptions($currOption) \
                        [lindex $options($currOption) 1]
                }
            }
        }
    }

    # Now process arguments

    set currArgNum 0

    while {$currArgNum != $numArgsDefined} {
        if {[lindex $args($currArgNum) 1 0] eq "REQUIRED"} {
            if {[llength $commandInput] == 0} {
                return [list [list 1 $errorBuffer] {} {}]
            }

            set outputArgs([lindex $args($currArgNum) 0]) \
                [lindex $commandInput 0]
            set commandInput [lrange $commandInput 1 end]
        } else {
            if {[llength $commandInput] != 0} {
                set outputArgs([lindex $args($currArgNum) 0]) \
                    [lindex $commandInput 0]
                set commandInput [lrange $commandInput 1 end]
            } else {
                set outputArgs([lindex $args($currArgNum) 0]) \
                    [lindex $args($currArgNum) 1 1]
            }
        }

        incr currArgNum
    }

    return [list [list 0 "Ok"] [array get outputOptions] [array get outputArgs]]
}


proc STAF::mapclassdef {args} {
    if {[llength $args] == 0} {
        error {wrong # args: should be "STAF::mapclassdef subcommand ?arg ...?"}
    }

    set subcommand [lindex $args 0]
    set args [lrange $args 1 end]

    switch $subcommand {
        create {
            set myArgs [list {name {}} {mapClassDefObj {}}]
            set checkResult [STAF::internalCheckCommandInput $args {} $myArgs]

            if {[lindex $checkResult 0 0] != 0} {
                error [lindex $checkResult 0 1]
            }

            array set input [lindex $checkResult 2]

            set mcd(name) $input(name)
            set mcd(keys) [list]

            return [array get mcd]
        }

        createInstance {
            # args: mapclassdefVar
            set myArgs [list mapclassdefVar]
            set checkResult [STAF::internalCheckCommandInput $args {} $myArgs]

            if {[lindex $checkResult 0 0] != 0} {
                error [lindex $checkResult 0 1]
            }

            array set input [lindex $checkResult 2]
            upvar $input(mapclassdefVar) mcd
            array set mcdArray $mcd

            set instance(staf-map-class-name) $mcdArray(name)

            return [array get instance]
        }

        addKey {
            # args: mapclassdefVar keyName ?displayName?
            set myArgs [list mapclassdefVar keyName {displayName {}}]
            set checkResult [STAF::internalCheckCommandInput $args {} $myArgs]

            if {[lindex $checkResult 0 0] != 0} {
                error [lindex $checkResult 0 1]
            }

            array set input [lindex $checkResult 2]
            upvar $input(mapclassdefVar) mcd

            set key(key) $input(keyName)
            set key(display-name) $input(displayName)

            array set mcdArray $mcd
            lappend mcdArray(keys) [array get key]
            set mcd [array get mcdArray]

            return $mcd
        }

        setKeyProperty {
            # args: mapclassdefVar, keyName, property, value):
            set myArgs [list mapclassdefVar keyName property value]
            set checkResult [STAF::internalCheckCommandInput $args {} $myArgs]

            if {[lindex $checkResult 0 0] != 0} {
                error [lindex $checkResult 0 1]
            }

            array set input [lindex $checkResult 2]
            upvar $input(mapclassdefVar) mcd

            array set mcdArray $mcd
            set outputKeys [list]

            foreach key $mcdArray(keys) {
                array set thisKey $key

                if {$thisKey(key) eq $input(keyName)} {
                    set thisKey($input(property)) $input(value)
                }

                lappend outputKeys [array get thisKey]
            }

            set mcdArray(keys) $outputKeys
            set mcd [array get mcdArray]

            return $mcd
        }

        keys {
            # args: mapclassdef
            set myArgs [list mapclassdef]
            set checkResult [STAF::internalCheckCommandInput $args {} $myArgs]

            if {[lindex $checkResult 0 0] != 0} {
                error [lindex $checkResult 0 1]
            }

            array set input [lindex $checkResult 2]
            array set mcdArray $input(mapclassdef)

            return $mcdArray(keys)
        }

        name {
            # args: mapclassdef
            set myArgs [list mapclassdef]
            set checkResult [STAF::internalCheckCommandInput $args {} $myArgs]

            if {[lindex $checkResult 0 0] != 0} {
                error [lindex $checkResult 0 1]
            }

            array set input [lindex $checkResult 2]
            array set mcdArray $input(mapclassdef)

            return $mcdArray(name)
        }

        getMapClassDefinitionObject {
            # args: mapclassdef
            set myArgs [list mapclassdef]
            set checkResult [STAF::internalCheckCommandInput $args {} $myArgs]

            if {[lindex $checkResult 0 0] != 0} {
                error [lindex $checkResult 0 1]
            }

            array set input [lindex $checkResult 2]

            return $input(mapclassdef)
        }
    }
}

proc STAF::mcontext {args} {
    if {[llength $args] == 0} {
        error {wrong # args: should be "STAF::mcontext subcommand ?arg ...?"}
    }

    set subcommand [lindex $args 0]
    set args [lrange $args 1 end]

    puts "subcommand: $subcommand"
    puts "args: $args"

    switch $subcommand {
        testOption {
            set myOptions [list opt1 {opt2 def2} opt3]
            set myArgs [list req1 req2 {oarg3 val3} {oarg4 val4}]
            set checkResult \
                [STAF::internalCheckCommandInput $args $myOptions $myArgs]

            if {[lindex $checkResult 0 0] != 0} {
                error [lindex $checkResult 0 1]
            }
        }

        create {
            set myArgs [list {rootObj {}} {mapClassMap {}}]
            set checkResult [STAF::internalCheckCommandInput $args {} $myArgs]

            if {[lindex $checkResult 0 0] != 0} {
                error [lindex $checkResult 0 1]
            }
            # array set options [lindex $checkResult 1]
            array set input [lindex $checkResult 2]

            return [array get input]
        }

        setMapClassDefinition {
            set myArgs [list mcVar mapClassDef]
            set checkResult [STAF::internalCheckCommandInput $args {} $myArgs]

            if {[lindex $checkResult 0 0] != 0} {
                error [lindex $checkResult 0 1]
            }

            array set input [lindex $checkResult 2]
            upvar $input(mcVar) mc
            array set mcArray $mc
            array set mcmArray $mcArray(mapClassMap)

            set mcdName [STAF::mapclassdef name $input(mapClassDef)]
            set mcmArray($mcdName) $input(mapClassDef)
            set mcArray(mapClassMap) [array get mcmArray]
            set mc [array get mcArray]

            return $mc
        }

        getMapClassDefinition {
            # args: mc mcdName
            set myArgs [list mc mcdName]
            set checkResult [STAF::internalCheckCommandInput $args {} $myArgs]

            if {[lindex $checkResult 0 0] != 0} {
                error [lindex $checkResult 0 1]
            }

            array set input [lindex $checkResult 2]
            array set mcArray $input(mc)
            array set mcmArray $mcArray(mapClassMap)

            if {[info exists mcmArray($mcdName)]} {
                return $mcmArray($mcdName)
            } else {
                return []
            }
        }

        hasMapClassDefinition {
            # args: mc mcdName
            set myArgs [list mc mcdName]
            set checkResult [STAF::internalCheckCommandInput $args {} $myArgs]

            if {[lindex $checkResult 0 0] != 0} {
                error [lindex $checkResult 0 1]
            }

            array set input [lindex $checkResult 2]
            array set mcArray $input(mc)
            array set mcmArray $mcArray(mapClassMap)

            return [info exists mcmArray($mcdName)]
        }

        getMapClassMap {
            # args: mc
            set myArgs [list mc]
            set checkResult [STAF::internalCheckCommandInput $args {} $myArgs]

            if {[lindex $checkResult 0 0] != 0} {
                error [lindex $checkResult 0 1]
            }

            array set input [lindex $checkResult 2]
            array set mcArray $input(mc)

            return $mcArray(mapClassMap)
        }

        mapClassDefinitionIterator {
            # args: mc
            set myArgs [list mc]
            set checkResult [STAF::internalCheckCommandInput $args {} $myArgs]

            if {[lindex $checkResult 0 0] != 0} {
                error [lindex $checkResult 0 1]
            }

            array set input [lindex $checkResult 2]
            array set mcArray $input(mc)
            array set mcmArray $mcArray(mapClassMap)

            return [array names mcmArray]
        }

        setRootObject {
            # args: mcVar rootObj
            set myArgs [list mcVar rootObj]
            set checkResult [STAF::internalCheckCommandInput $args {} $myArgs]

            if {[lindex $checkResult 0 0] != 0} {
                error [lindex $checkResult 0 1]
            }

            array set input [lindex $checkResult 2]
            upvar $input(mcVar) mc
            array set mcArray $mc
            set mcArray(rootObj) $input(rootObj)
            set mc [array get mcArray]

            return $mc
        }

        getRootObject {
            # args: mc
            set myArgs [list mc]
            set checkResult [STAF::internalCheckCommandInput $args {} $myArgs]

            if {[lindex $checkResult 0 0] != 0} {
                error [lindex $checkResult 0 1]
            }

            array set input [lindex $checkResult 2]
            array set mcArray $input(mc)

            return $mcArray(rootObject)
        }

        getPrimaryObject {
            # args: mc
            set myArgs [list mc]
            set checkResult [STAF::internalCheckCommandInput $args {} $myArgs]

            if {[lindex $checkResult 0 0] != 0} {
                error [lindex $checkResult 0 1]
            }

            array set input [lindex $checkResult 2]
            array set mcArray $input(mc)
            array set mcmArray $mcArray(mapClassMap)

            if {[array size mcmArray]} {
                return $mc
            } else {
                return $mcArray(rootObj)
            }
        }

        marshall {
            # self
            puts "Not defined yet"
        }

        unmarshall {
            puts "Not defined yet"
        }
    }
}


proc STAF::unmarshall {data {context {}} {flags {}}} {
}