/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

package com.ibm.staf.service.stax;
import org.w3c.dom.Node;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.NodeList;

public class STAXLogActionFactory implements STAXActionFactory
{
    private static String fDTDInfo =
"\n" +
"<!--================= The Log Element ============================== -->\n" +
"<!--\n" +
"     Writes a message and its log level to a STAX Job User Log file.\n" +
"     The message must evaluate via Python to a string.\n\n" +
"     The log level specified defaults to 'info'.  If specified, it\n" +
"     must evaluate via Python to a string containing one of the\n" +
"     following STAF Log Service Log levels:\n" +
"       fatal, warning, info, trace, trace2, trace3, debug, debug2,\n" +
"       debug3, start, stop, pass, fail, status, user1, user2, user3,\n" +
"       user4, user5, user6, user7, user8\n" +
"     The message attribute is evaluated via Python.  If it evaluates\n" +
"     to true, the message text will also be sent to the STAX Job Monitor.\n" +
"     The message attribute defaults to the STAXMessageLog variable whose\n" +
"     value defaults to 0 (false) but can by changed within the STAX job\n" +
"     to turn on messaging.\n\n" +
"     If an if attribute is specified and it evaluates via Python to\n" +
"     false, then the log element is ignored.\n" +  
"-->\n" +
"<!ELEMENT log         (#PCDATA)>\n" +
"<!ATTLIST log\n" +
"          level       CDATA       \"'info'\"\n" +
"          message     CDATA       \"STAXMessageLog\"\n" +
"          if          CDATA       \"1\"\n" +
">\n";

    public String getDTDInfo()
    {
        return fDTDInfo;
    }

    public String getDTDTaskName()
    {
        return "log";
    }

    public STAXAction parseAction(STAX staxService, STAXJob job,
                                  org.w3c.dom.Node root) throws STAXException
    {
        STAXLogAction action = new STAXLogAction();
        action.setLogfile(STAXJob.USER_JOB_LOG);
        
        NamedNodeMap attrs = root.getAttributes();

        for (int i = 0; i < attrs.getLength(); ++i)
        {
            Node thisAttr = attrs.item(i);

            String errorInfo = "\n  Element: " + root.getNodeName() +
                               "  Attribute: " + thisAttr.getNodeName();

            if (thisAttr.getNodeName().equals("level"))
            {
                if (thisAttr.getNodeValue().equals(""))
                {
                    throw new STAXInvalidXMLAttributeException(
                        errorInfo + "\nThe value for the level attribute " +
                        "cannot be blank");
                }

                action.setLevel(STAXUtil.parseAndCompileForPython(
                    thisAttr.getNodeValue(), errorInfo));
            }
            else if (thisAttr.getNodeName().equals("message"))
            {
                action.setMessageAttr(STAXUtil.parseAndCompileForPython(
                    thisAttr.getNodeValue(), errorInfo));
            }
            else if (thisAttr.getNodeName().equals("if"))
            {
                action.setIf(STAXUtil.parseAndCompileForPython(
                    thisAttr.getNodeValue(), errorInfo));
            }
            else
            {
                throw new STAXInvalidXMLAttributeException(
                    root.getNodeName() + ": " + thisAttr.getNodeName());
            }
        }

        String errorInfo = "\n  Element: " + root.getNodeName();

        NodeList children = root.getChildNodes();

        for (int i = 0; i < children.getLength(); ++i)
        {
            Node thisChild = children.item(i);

            if (thisChild.getNodeType() == Node.COMMENT_NODE)
            {
                /* Do nothing */
            }
            else if (thisChild.getNodeType() == Node.TEXT_NODE)
            {
                action.setMessage(STAXUtil.parseAndCompileForPython(
                    thisChild.getNodeValue(), errorInfo));
            }
            else
            {
                throw new STAXInvalidXMLNodeTypeException(
                    Integer.toString(thisChild.getNodeType()));
            }
        }

        return action;
    }
}
