/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include <map>
#include "STAFFileSystem.h"
#include "STAFString.h"
#include "STAFMutexSem.h"
#include "STAFRWSem.h"
#include "STAF_fstream.h"

struct FileLock
{
    FileLock() : lock(0), lockSem(new STAFMutexSem, STAFMutexSemPtr::INIT),
                 numOwners(0), rwSem(new STAFRWSem, STAFRWSemPtr::INIT)
    { /* Do Nothing */ }

    STAFFSOSFileLock_t lock;
    STAFMutexSemPtr lockSem;
    unsigned int numOwners;
    STAFRWSemPtr rwSem;
};

typedef std::map<STAFString, FileLock> LockMap;

static LockMap sLockMap;
static STAFMutexSem sLockMapSem;
static STAFMutexSem sCurrDirSem;
static STAFString sSemiColon(kUTF8_SCOLON);


/*****************************************************************************/
/* STAFFSStringMatchesWildcards - Determines whether a given string matches  */
/*                                a given wildcard string                    */
/*                                                                           */
/* Accepts: (In)  String to check                                            */
/*          (In)  String containg wildcards                                  */
/*          (In)  Case sensitivity indicator                                 */
/*          (Out) Result of comparison (0 = no match, 1 = matches)           */
/*                                                                           */
/* Returns: Standard return codes                                            */
/*                                                                           */
/* Notes  : 1) Two wildcard characters are understood.  '*' matches zero or  */
/*             more characters.  '?' matches one single character.           */
/*          2) Doesn't do any backtracking, so for example, if the           */
/*             stringToCheck is XYYXYXY and the wildcardString is ?Y*X?,     */
/*             they will not match, but strings XYYYXY and XYYYXX will.      */
/*****************************************************************************/
STAFRC_t STAFFSStringMatchesWildcards(STAFStringConst_t stringToCheck,
                                      STAFStringConst_t wildcardString,
                                      STAFFSCaseSensitive_t sensitive,
                                      unsigned int *matches)
{
    if (sensitive == kSTAFFSCaseDefault)
        STAFFSInfo(&sensitive, kSTAFFSCaseSensitivity);

    STAFStringCaseSensitive_t caseSensitive;

    if (sensitive == kSTAFFSCaseInsensitive)
        caseSensitive = kSTAFStringCaseInsensitive;
    else
        caseSensitive = kSTAFStringCaseSensitive;

    return STAFStringMatchesWildcards(stringToCheck, wildcardString, 
                                      caseSensitive, matches, 0);
}


// XXX: Should provide a common routine for the Read/WriteLock and
//      Read/WriteUnlock APIs.  They only differ by one line of code

STAFRC_t STAFFSEntryReadLock(STAFFSEntry_t entry, unsigned int *osRC)
{
    if (entry == 0) return kSTAFInvalidObject;

    STAFRC_t retCode = kSTAFOk;

    try
    {
        STAFStringConst_t filenameT = 0;
        STAFRC_t rc = STAFFSEntryGetPathString(entry, &filenameT, osRC);

        if (rc) return rc;

        STAFString filename(filenameT);
        FileLock *pLock = 0;

        // First get the lock object from the map

        {
            STAFMutexSemLock mapLock(sLockMapSem);
            pLock = &sLockMap[filename];
        }

        // Now, get the underlying file system lock if needed, and increment
        // the number of owners

        {
            STAFMutexSemLock lockLock(*pLock->lockSem);

            if (pLock->lock == 0)
            {
                STAFRC_t rc = STAFFSOSGetExclusiveFileLock(filenameT,
                                                           &pLock->lock, osRC);
                if (rc != 0) return rc;
            }

            ++pLock->numOwners;
        }

        // Finally, get a read lock

        pLock->rwSem->readLock();
    }
    catch (...)
    { retCode = kSTAFUnknownError; }

    return retCode;
}


STAFRC_t STAFFSEntryReadUnlock(STAFFSEntry_t entry, unsigned int *osRC)
{
    if (entry == 0) return kSTAFInvalidObject;

    STAFRC_t retCode = kSTAFOk;

    try
    {
        STAFStringConst_t filenameT = 0;
        STAFRC_t rc = STAFFSEntryGetPathString(entry, &filenameT, osRC);

        if (rc) return rc;

        STAFString filename(filenameT);
        FileLock *pLock = 0;

        // First get the lock object from the map

        {
            STAFMutexSemLock mapLock(sLockMapSem);
            pLock = &sLockMap[filename];
        }

        // Now, decrement the number of owners and release the underlying file
        // system lock if necessary

        STAFMutexSemLock lockLock(*pLock->lockSem);

        if (--pLock->numOwners == 0)
        {
            STAFRC_t rc = STAFFSOSReleaseExclusiveFileLock(&pLock->lock, osRC);

            if (rc != 0) return rc;

            pLock->lock = 0;
        }

        // Finally, release the read lock

        pLock->rwSem->readUnlock();
    }
    catch (...)
    { retCode = kSTAFUnknownError; }

    return retCode;
}


STAFRC_t STAFFSEntryWriteLock(STAFFSEntry_t entry, unsigned int *osRC)
{
    if (entry == 0) return kSTAFInvalidObject;

    STAFRC_t retCode = kSTAFOk;

    try
    {
        STAFStringConst_t filenameT = 0;
        STAFRC_t rc = STAFFSEntryGetPathString(entry, &filenameT, osRC);

        if (rc) return rc;

        STAFString filename(filenameT);
        FileLock *pLock = 0;

        // First get the lock object from the map

        {
            STAFMutexSemLock mapLock(sLockMapSem);
            pLock = &sLockMap[filename];
        }

        // Now, get the underlying file system lock if needed, and increment
        // the number of owners

        {
            STAFMutexSemLock lockLock(*pLock->lockSem);

            if (pLock->lock == 0)
            {
                STAFRC_t rc = STAFFSOSGetExclusiveFileLock(filenameT,
                                                           &pLock->lock, osRC);
                if (rc != 0) return rc;
            }

            ++pLock->numOwners;
        }

        // Finally, get a write lock

        pLock->rwSem->writeLock();
    }
    catch (...)
    { retCode = kSTAFUnknownError; }

    return retCode;
}


STAFRC_t STAFFSEntryWriteUnlock(STAFFSEntry_t entry, unsigned int *osRC)
{
    if (entry == 0) return kSTAFInvalidObject;

    STAFRC_t retCode = kSTAFOk;

    try
    {
        STAFStringConst_t filenameT = 0;
        STAFRC_t rc = STAFFSEntryGetPathString(entry, &filenameT, osRC);

        if (rc) return rc;

        STAFString filename(filenameT);
        FileLock *pLock = 0;

        // First get the lock object from the map

        {
            STAFMutexSemLock mapLock(sLockMapSem);
            pLock = &sLockMap[filename];
        }

        // Now, decrement the number of owners and release the underlying file
        // system lock if necessary

        STAFMutexSemLock lockLock(*pLock->lockSem);

        if (--pLock->numOwners == 0)
        {
            STAFRC_t rc = STAFFSOSReleaseExclusiveFileLock(&pLock->lock, osRC);

            if (rc != 0) return rc;

            pLock->lock = 0;
        }

        // Finally, release the write lock

        pLock->rwSem->writeUnlock();
    }
    catch (...)
    { retCode = kSTAFUnknownError; }

    return retCode;
}


STAFRC_t STAFFSRequestCurrentDirectoryLock()
{
    STAFRC_t retCode = kSTAFOk;

    try
    {
        sCurrDirSem.request();
    }
    catch (...)
    { retCode = kSTAFUnknownError; }

    return retCode;
}


STAFRC_t STAFFSReleaseCurrentDirectoryLock()
{
    STAFRC_t retCode = kSTAFOk;

    try
    {
        sCurrDirSem.release();
    }
    catch (...)
    { retCode = kSTAFUnknownError; }

    return retCode;
}


STAFRC_t STAFFSCopyEntryCommon(STAFFSEntry_t source, STAFStringConst_t target,
                               unsigned int *osRC)
{
    if (source == 0) return kSTAFInvalidObject;
    if (target == 0) return kSTAFInvalidParm;

    STAFRC_t retCode = kSTAFOk;

    try
    {
        STAFStringConst_t filenameT = 0;
        STAFRC_t rc = STAFFSEntryGetPathString(source, &filenameT, osRC);

        if (rc) return rc;

        STAFString filename(filenameT);

        fstream infile(STAFString(filenameT).toCurrentCodePage()->buffer(),
                       ios::in | STAF_ios_binary);
        fstream outfile(STAFString(target).toCurrentCodePage()->buffer(),
                        ios::out | STAF_ios_binary | ios::trunc);

        if (!infile || !outfile) return kSTAFFileOpenError;

        char input = 0;

        while (infile.read(&input, 1)) outfile.write(&input, 1);
    }
    catch (...)
    { retCode = kSTAFUnknownError; }

    return retCode;
}
