/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAF_iostream.h"
#include "STAFThreadManager.h"
#include "STAFException.h"


STAFThreadManager::STAFThreadManager(unsigned int initThreads,
                                     unsigned int growthDelta)
  : fGrowthDelta(growthDelta), fThreadPoolSize(0),
    fCurrReadyThread(0)
{
    doGrowThreadPool(initThreads);
}


STAFThreadID_t STAFThreadManager::getCurrentThreadID()
{
    return STAFThreadCurrentThreadID();
}


void STAFThreadManager::sleepCurrentThread(unsigned int milliseconds)
{
    unsigned int osRC = 0;

    STAFThreadSleepCurrentThread(milliseconds, &osRC);
}


unsigned int STAFThreadManager::dispatch(STAFThreadFunc_t func, void *data)
{
    STAFMutexSemLock semLock(fThreadPoolSem);

    if (fThreadPool.size() == 0)
    {
        // Need to grow pool
        if (!fGrowthDelta) return 1;

        doGrowThreadPool();
    }

    fCurrReadyThread = fThreadPool.back();
    fThreadPool.pop_back();

    fCurrReadyThread->work = new STAFThreadFunc(func, data);
    fCurrReadyThread->workAvailable.post();

    return 0;
}

unsigned int STAFThreadManager::getThreadPoolSize()
{
    return fThreadPoolSize;
}


unsigned int STAFThreadManager::getNumReadyThreads()
{
    STAFMutexSemLock semLock(fThreadPoolSem);

    return fThreadPool.size();
}


unsigned int STAFThreadManager::getNumWorkingThreads()
{
    STAFMutexSemLock semLock(fThreadPoolSem);

    return fThreadPoolSize - fThreadPool.size();
}


unsigned int STAFThreadManager::getGrowthDelta()
{
    return fGrowthDelta;
}


void STAFThreadManager::setGrowthDelta(unsigned int growthDelta)
{
    fGrowthDelta = growthDelta;
}


void STAFThreadManager::growThreadPool()
{
    growThreadPool(fGrowthDelta);
}

void STAFThreadManager::growThreadPool(unsigned int deltaThreads)
{
    STAFMutexSemLock semLock(fThreadPoolSem);
    doGrowThreadPool(deltaThreads);
}


void STAFThreadManager::doGrowThreadPool()
{
    doGrowThreadPool(fGrowthDelta);
}


void STAFThreadManager::doGrowThreadPool(unsigned int deltaThreads)
{
    for(int i = 0; i < deltaThreads; ++i)
    {
        fCurrReadyThread = new STAFReadyThread();
        fWorkerSynchSem.reset();

        // Note: STAFThreadID_t is not a simple integer type on z/OS, thus
        //       we can't initialize theThreadID

        STAFThreadID_t theThreadID;
        unsigned int osRC = 0;
        unsigned int rc = STAFThreadStart(&theThreadID, callWorkerThread,
                                          this, 0, &osRC);

        // XXX: Do something on error
        fWorkerSynchSem.wait();
        fThreadPool.push_back(fCurrReadyThread);
        ++fThreadPoolSize;
    }
}

unsigned int STAFThreadManager::callWorkerThread(void *manager)
{
     STAFThreadManager *threadManager =
         static_cast<STAFThreadManager *>(manager);

     threadManager->workerThread();

     return 0;
}


void STAFThreadManager::workerThread()
{
    try
    {
        STAFReadyThread *myThreadData = fCurrReadyThread;
        fWorkerSynchSem.post();

        for(;;)
        {
            myThreadData->workAvailable.wait();
            myThreadData->workAvailable.reset();

            if (myThreadData->work == 0) break;

            try
            {
                myThreadData->work->func(myThreadData->work->data);
                delete myThreadData->work;
            }
            catch (STAFException &se)
            { se.write("In STAFThreadManager::workerThread()"); }
            catch (...)
            {
                cout << "Caught unknown exception in "
                     << "STAFThreadManager::workerThread()" << endl;
            }

            // Add the thread back into the thread pool

            STAFMutexSemLock semLock(fThreadPoolSem);
            fThreadPool.push_back(myThreadData);
        }

        delete myThreadData;
    }
    catch (STAFException &se)
    { se.write("In STAFThreadManager::workerThread() outer try block" ); }
    catch (...)
    {
        cout << "Caught unknown exception in "
             << "STAFThreadManager::workerThread() outer try block" << endl;
    }
}


STAFThreadManager::~STAFThreadManager()
{
    // Simply have all ready threads end
    STAFMutexSemLock semLock(fThreadPoolSem);

    while (fThreadPool.size() > 0)
    {
        fCurrReadyThread = fThreadPool.back();
        fThreadPool.pop_back();

        fCurrReadyThread->work = 0;
        fCurrReadyThread->workAvailable.post();
    }
}
