/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001, 2004, 2005                                  */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAFProc.h"
#include "STAFProcUtil.h"
#include "STAFSemService.h"
#include "STAFCommandParser.h"
#include "STAFHandleService.h"
#include "STAFHandleManager.h"
#include "STAFUtil.h"

STAFHandleService *handleService;

STAFSemService::STAFSemService() : STAFService("SEM")
{
    // Set up Request Parser

    fRequestParser.addOption("REQUEST", 1,
                             STAFCommandParser::kValueNotAllowed);
    fRequestParser.addOption("MUTEX", 1,
                             STAFCommandParser::kValueRequired);
    fRequestParser.addOption("TIMEOUT", 1,
                             STAFCommandParser::kValueRequired);
    fRequestParser.addOptionGroup("MUTEX", 1, 1);
    
    // Set up Release Parser

    fReleaseParser.addOption("RELEASE", 1,
                             STAFCommandParser::kValueNotAllowed);
    fReleaseParser.addOption("MUTEX", 1,
                             STAFCommandParser::kValueRequired);
    fReleaseParser.addOption("FORCE", 1,
                             STAFCommandParser::kValueNotAllowed);
    fRequestParser.addOptionGroup("MUTEX", 1, 1);

    // Set up Post/Pulse Parser

    fPostPulseParser.addOption("POST", 1,
                               STAFCommandParser::kValueNotAllowed);
    fPostPulseParser.addOption("PULSE", 1,
                               STAFCommandParser::kValueNotAllowed);
    fPostPulseParser.addOption("EVENT", 1,
                               STAFCommandParser::kValueRequired);
    fPostPulseParser.addOptionGroup("EVENT", 1, 1);
    fPostPulseParser.addOptionGroup("POST PULSE", 1, 1);

    // Set up Reset Parser

    fResetParser.addOption("RESET", 1,
                           STAFCommandParser::kValueNotAllowed);
    fResetParser.addOption("EVENT", 1,
                           STAFCommandParser::kValueRequired);
    fResetParser.addOptionGroup("EVENT", 1, 1);

    // Set up Wait Parser

    fWaitParser.addOption("WAIT", 1,
                          STAFCommandParser::kValueNotAllowed);
    fWaitParser.addOption("EVENT", 1,
                          STAFCommandParser::kValueRequired);
    fWaitParser.addOption("TIMEOUT", 1,
                          STAFCommandParser::kValueRequired);
    fWaitParser.addOptionGroup("EVENT", 1, 1);

    // Delete Parser Setup
 
    fDeleteParser.addOption("DELETE",  1, 
                            STAFCommandParser::kValueNotAllowed);
    fDeleteParser.addOption("MUTEX", 1, 
                            STAFCommandParser::kValueRequired);
    fDeleteParser.addOption("EVENT", 1, 
                            STAFCommandParser::kValueRequired);
    fDeleteParser.addOptionGroup("MUTEX EVENT", 1, 1);

    // Query Parser Setup
 
    fQueryParser.addOption("QUERY",  1, 
                           STAFCommandParser::kValueNotAllowed);
    fQueryParser.addOption("MUTEX", 1, 
                           STAFCommandParser::kValueRequired);
    fQueryParser.addOption("EVENT", 1, 
                           STAFCommandParser::kValueRequired);
    fQueryParser.addOptionGroup("MUTEX EVENT", 1, 1);

    // List Parser Setup
 
    fListParser.addOption("LIST",  1, 
        STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("MUTEX", 1, 
        STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("EVENT", 1, 
        STAFCommandParser::kValueNotAllowed);
    fListParser.addOptionGroup("MUTEX EVENT", 1, 1);
        
    // staf_callback Parser Setup
 
    fSTAFCallbackParser.addOption("STAF_CALLBACK",  1, 
        STAFCommandParser::kValueNotAllowed);
    fSTAFCallbackParser.addOption("HANDLEDELETED", 1, 
        STAFCommandParser::kValueNotAllowed);
    fSTAFCallbackParser.addOption("HANDLE", 1, 
        STAFCommandParser::kValueRequired);
    fSTAFCallbackParser.addOption("MACHINE", 1, 
        STAFCommandParser::kValueRequired);
    fSTAFCallbackParser.addOption("UUID", 1, 
        STAFCommandParser::kValueRequired);
    fSTAFCallbackParser.addOption("KEY", 1, 
        STAFCommandParser::kValueRequired);
    
    handleService = STAFHandleService::getInstance();
    
    // Construct map class for listing mutex semaphore information

    fMutexInfoClass = STAFMapClassDefinition::create(
        "STAF/Service/Sem/MutexInfo");
 
    fMutexInfoClass->addKey("name", "Name");
    fMutexInfoClass->addKey("state", "State");
    fMutexInfoClass->addKey("pendingRequests", "Pending Requests");

    // Construct map class for listing event semaphore information

    fEventInfoClass = STAFMapClassDefinition::create(
        "STAF/Service/Sem/EventInfo");
 
    fEventInfoClass->addKey("name", "Name");
    fEventInfoClass->addKey("state", "State");
    fEventInfoClass->addKey("waiters", "Waiters");

    // Construct map class for detailed information on an event semaphore

    fQueryEventClass = STAFMapClassDefinition::create(
        "STAF/Service/Sem/QueryEvent");
 
    fQueryEventClass->addKey("state", "State");
    fQueryEventClass->addKey("lastPosted", "Last Posted");
    fQueryEventClass->addKey("lastReset", "Last Reset");
    fQueryEventClass->addKey("waiterList", "Waiters");
    
    // Construct map class for detailed information on a mutex semaphore

    fQueryMutexClass = STAFMapClassDefinition::create(
        "STAF/Service/Sem/QueryMutex");
 
    fQueryMutexClass->addKey("state", "State");
    fQueryMutexClass->addKey("owner", "Owner");
    fQueryMutexClass->addKey("requestList", "Pending Requests");

    // Construct map class for an event semaphore requester

    fEventRequesterClass = STAFMapClassDefinition::create(
        "STAF/Service/Sem/EventRequester");

    fEventRequesterClass->addKey("machine", "Machine");
    fEventRequesterClass->addKey("handleName", "Handle Name");
    fEventRequesterClass->addKey("handle", "Handle");
    fEventRequesterClass->addKey("user", "User");
    fEventRequesterClass->addKey("endpoint", "Endpoint");
    fEventRequesterClass->addKey("timestamp", "Date-Time");

    // Construct map class for an event semaphore requester

    fMutexOwnerClass = STAFMapClassDefinition::create(
        "STAF/Service/Sem/MutexOwner");

    fMutexOwnerClass->addKey("machine", "Machine");
    fMutexOwnerClass->addKey("handleName", "Handle Name");
    fMutexOwnerClass->addKey("handle", "Handle");
    fMutexOwnerClass->addKey("user", "User");
    fMutexOwnerClass->addKey("endpoint", "Endpoint");
    fMutexOwnerClass->addKey("requestTimestamp", "Date-Time Requested");
    fMutexOwnerClass->addKey("acquireTimestamp", "Date-Time Acquired");

    // Construct map class for an mutex semaphore requester

    fPendingRequestClass = STAFMapClassDefinition::create(
        "STAF/Service/Sem/PendingRequest");

    fPendingRequestClass->addKey("machine", "Machine");
    fPendingRequestClass->addKey("handleName", "Handle Name");
    fPendingRequestClass->addKey("handle", "Handle");
    fPendingRequestClass->addKey("user", "User");
    fPendingRequestClass->addKey("endpoint", "Endpoint");
    fPendingRequestClass->addKey("requestTimestamp", "Date-Time Requested");
}


STAFSemService::~STAFSemService()
{
    /* Do Nothing */
}


STAFString STAFSemService::info(unsigned int) const
{
    return name() + ": Internal";
}


STAFServiceResult STAFSemService::acceptRequest(
    const STAFServiceRequest &requestInfo)
{
    STAFString action = requestInfo.fRequest.subWord(0, 1).lowerCase();
 
    if      (action == "request") return handleRequest(requestInfo);
    else if (action == "release") return handleRelease(requestInfo);
    else if (action == "post" || action == "pulse")
        return handlePostPulse(requestInfo);
    else if (action == "reset") return handleReset(requestInfo);
    else if (action == "wait") return handleWait(requestInfo);
    else if (action == "delete") return handleDelete(requestInfo);
    else if (action == "query") return handleQuery(requestInfo);
    else if (action == "list")  return handleList(requestInfo);
    else if (action == "staf_callback") return handleSTAFCallback(requestInfo);
    else if (action == "help")  return handleHelp(requestInfo);
    else return kSTAFInvalidRequestString;
}


STAFServiceResult STAFSemService::handleRequest(
    const STAFServiceRequest &requestInfo)
{
    // Verify that the requesting machine/user has at least trust level 3

    IVALIDATE_TRUST(3, "REQUEST");

    // Parse the request

    STAFCommandParseResultPtr parsedResult =
        fRequestParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }

    DEFINE_VAR_POOL_LIST(varPoolList, varPoolListSize, requestInfo);
    STAFString errorBuffer;
    STAFString result;
    STAFString semName;
    STAFRC_t rc = RESOLVE_STRING_OPTION("MUTEX", semName);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    STAFString uppercaseSemName = semName.toUpperCase();

    unsigned int timeout = STAF_EVENT_SEM_INDEFINITE_WAIT;
    
    rc = RESOLVE_DEFAULT_UINT_OPTION("TIMEOUT", timeout,
                                     STAF_EVENT_SEM_INDEFINITE_WAIT);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    fMutexSemListSem.request();

    // GCH
    // Register for notification when the handle ends
  
    gHandleManagerPtr->addNotification(requestInfo.fHandle,
        requestInfo.fMachine, requestInfo.fSTAFInstanceUUID, 
        name(), "SemServiceMutex");

    if (fMutexSemList.find(uppercaseSemName) != fMutexSemList.end())
    {
        // The sempahore exists.  Get it and lock it.

        MutexSemPtr sem = fMutexSemList[uppercaseSemName];

        sem->lockPtr->request();

        if (sem->isOwned)
        {
            // It is safe to unlock the semaphore list here, as this
            // semaphore can't be deleted with requests pending

            fMutexSemListSem.release();

            // Add a request to the list, unlock the semaphore,
            // and wait to acquire ownership

            MutexRequest request(requestInfo.fSTAFInstanceUUID,
                                 requestInfo.fMachine,
                                 requestInfo.fHandleName,
                                 requestInfo.fHandle,
                                 requestInfo.fUser,
                                 requestInfo.fEndpoint);
            STAFEventSemPtr myTurn = request.avail;

            sem->requestList.push_back(request);
            sem->lockPtr->release();

            myTurn->wait(timeout);

            sem->lockPtr->request();

            if (myTurn->query() == kSTAFEventSemPosted)
            {
                // We were posted.  So, we should have been given
                // ownership.  Yeah.

                sem->lockPtr->release();
            }
            else
            {
                // We timed out waiting for the semaphore, so remove
                // us from the list

                int foundUs = 0;
                MutexRequestList::iterator iter;

                for(iter = sem->requestList.begin(); !foundUs && 
                    iter != sem->requestList.end(); iter++)
                {
                    if ((iter->stafInstanceUUID ==
                         requestInfo.fSTAFInstanceUUID) &&
                        (iter->handle == requestInfo.fHandle))
                    {
                        foundUs = 1;
                        sem->requestList.erase(iter);
                        break;
                    }
                }

                // Now unlock the semaphore and return a timeout error

                sem->lockPtr->release();

                return kSTAFTimeout;
            }
        }
        else
        {
            // Take ownership of the semaphore

            sem->isOwned = 1;
            sem->owner = MutexRequest(requestInfo.fSTAFInstanceUUID,
                                      requestInfo.fMachine,
                                      requestInfo.fHandleName,
                                      requestInfo.fHandle,
                                      requestInfo.fUser,
                                      requestInfo.fEndpoint);
            sem->acquireTimeDate = STAFTimestamp::now();

            // Now release the locks on the semphore and the list

            fMutexSemListSem.release();
            sem->lockPtr->release();
        }
    }
    else
    {
        // The semaphore doesn't exist, so create it.

        fMutexSemList[uppercaseSemName] =
            MutexSemPtr(new MutexSem(semName, 1, 
                                     MutexRequest(
                                         requestInfo.fSTAFInstanceUUID,
                                         requestInfo.fMachine,
                                         requestInfo.fHandleName,
                                         requestInfo.fHandle,
                                         requestInfo.fUser,
                                         requestInfo.fEndpoint)),
                        MutexSemPtr::INIT);

        // Now release the lock on the list

        fMutexSemListSem.release();
    }

    return STAFServiceResult(kSTAFOk, result);
}

STAFServiceResult STAFSemService::handleRelease(
    const STAFServiceRequest &requestInfo)
{
    STAFCommandParseResultPtr parsedResult =
        fReleaseParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }

    DEFINE_VAR_POOL_LIST(varPoolList, varPoolListSize, requestInfo);
    STAFString errorBuffer;
    STAFString result;
    STAFString semName;
    STAFRC_t rc = RESOLVE_STRING_OPTION("MUTEX", semName);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    STAFString uppercaseSemName = semName.toUpperCase();

    // Verify that the requesting machine/user has at least trust level 3 if
    // FORCE is not specified and at least trust level 4 if FORCE is specified

    if (parsedResult->optionTimes("FORCE") == 0)
    {
        IVALIDATE_TRUST(3, "RELEASE");
    }
    else
    {
        IVALIDATE_TRUST(4, "RELEASE FORCE");
    }
    
    fMutexSemListSem.request();

    if (fMutexSemList.find(uppercaseSemName) == fMutexSemList.end())
    {
        fMutexSemListSem.release();
        return kSTAFSemaphoreDoesNotExist;
    }

    // Ok, the semaphore exists.  Get it and lock it, then release the
    // lock on the semaphore list.

    MutexSemPtr sem = fMutexSemList[uppercaseSemName];
    STAFMutexSemLock semLock(*sem->lockPtr);

    fMutexSemListSem.release();

    // Verify we are allowed to release the semaphore

    if (((requestInfo.fSTAFInstanceUUID != sem->owner.stafInstanceUUID) ||
        (requestInfo.fHandle != sem->owner.handle)) &&
        (parsedResult->optionTimes("FORCE") == 0))
    {
        return kSTAFNotSemaphoreOwner;
    }

    gHandleManagerPtr->deleteNotification(sem->owner.handle, 
                                          sem->owner.machine,
                                          sem->owner.stafInstanceUUID,
                                          name());
        
    if (sem->requestList.size() != 0)
    {
        // There is a pending request.  Remove it from the list and
        // notify the waiter.

        sem->acquireTimeDate = STAFTimestamp::now();

        sem->owner = sem->requestList.front();
        sem->requestList.pop_front();

        sem->owner.avail->post();
    }
    else
    {
        sem->isOwned = 0;
    }

    return STAFServiceResult(kSTAFOk, result);
}

STAFServiceResult STAFSemService::handlePostPulse(
    const STAFServiceRequest &requestInfo)
{
    // Parse the request

    STAFCommandParseResultPtr parsedResult =
        fPostPulseParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }                             

    // Verify that the requesting machine/user has at least trust level 3
    
    if (parsedResult->optionTimes("POST") != 0)
    {
        IVALIDATE_TRUST(3, "POST");
    }
    else
    {
        IVALIDATE_TRUST(3, "PULSE");
    }

    DEFINE_VAR_POOL_LIST(varPoolList, varPoolListSize, requestInfo);
    STAFString errorBuffer;
    STAFString result;
    STAFString semName;
    STAFRC_t rc = RESOLVE_STRING_OPTION("EVENT", semName);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    STAFString uppercaseSemName = semName.toUpperCase();

    fEventSemListSem.request();

    if (fEventSemList.find(uppercaseSemName) == fEventSemList.end())
    {
        // Semaphore does not exist, so create it in a posted state
        // if on POST, or in a reset state if on PULSE

        EventSemPtr sem = EventSemPtr(new EventSem(semName), 
                                      EventSemPtr::INIT);

        sem->lastPost = EventUnit(requestInfo.fMachine,
                                  requestInfo.fHandleName,
                                  requestInfo.fHandle,
                                  requestInfo.fUser,
                                  requestInfo.fEndpoint);

        if (parsedResult->optionTimes("POST") != 0)
        {
            sem->eventSem->post();
        }
        else
        {
            sem->eventSem->reset();
            sem->lastReset = sem->lastPost;
        }

        fEventSemList[uppercaseSemName] = sem;

        fEventSemListSem.release();
    }
    else
    {
        // The semaphore does exist, so lock the sem, unlock the main
        // list, clear the waiter list, and post/pulse it.

        EventSemPtr sem = fEventSemList[uppercaseSemName];

        sem->lockPtr->request();
        fEventSemListSem.release();

        sem->waiterList.clear();
        sem->lastPost = EventUnit(requestInfo.fMachine,
                                  requestInfo.fHandleName,
                                  requestInfo.fHandle,
                                  requestInfo.fUser,
                                  requestInfo.fEndpoint);
        sem->eventSem->post();

        if (parsedResult->optionTimes("PULSE") != 0)
        {
            sem->eventSem->reset();
            sem->lastReset = sem->lastPost;
        }

        sem->lockPtr->release();
    }

    return STAFServiceResult(kSTAFOk, result);
}

STAFServiceResult STAFSemService::handleReset(
    const STAFServiceRequest &requestInfo)
{
    // Verify that the requesting machine/user has at least trust level 3

    IVALIDATE_TRUST(3, "RESET");

    // Parse the request

    STAFCommandParseResultPtr parsedResult =
        fResetParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }                             

    DEFINE_VAR_POOL_LIST(varPoolList, varPoolListSize, requestInfo);
    STAFString errorBuffer;
    STAFString result;
    STAFString semName;
    STAFRC_t rc = RESOLVE_STRING_OPTION("EVENT", semName);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    STAFString uppercaseSemName = semName.toUpperCase();

    fEventSemListSem.request();

    if (fEventSemList.find(uppercaseSemName) == fEventSemList.end())
    {
        // Semaphore does not exist, so create it in a reset state

        EventSemPtr sem = EventSemPtr(new EventSem(semName), 
                                      EventSemPtr::INIT);

        sem->lastReset = EventUnit(requestInfo.fMachine,
                                   requestInfo.fHandleName,
                                   requestInfo.fHandle,
                                   requestInfo.fUser,
                                   requestInfo.fEndpoint);
        sem->eventSem->reset();

        fEventSemList[uppercaseSemName] = sem;

        fEventSemListSem.release();
    }
    else
    {
        // The semaphore does exist, so lock the sem, unlock the main
        // list, and reset it.

        EventSemPtr sem = fEventSemList[uppercaseSemName];

        sem->lockPtr->request();
        fEventSemListSem.release();

        sem->lastReset = EventUnit(requestInfo.fMachine,
                                   requestInfo.fHandleName,
                                   requestInfo.fHandle,
                                   requestInfo.fUser,
                                   requestInfo.fEndpoint);
        sem->eventSem->reset();

        sem->lockPtr->release();
    }

    return STAFServiceResult(kSTAFOk, result);
}

STAFServiceResult STAFSemService::handleWait(
    const STAFServiceRequest &requestInfo)
{
    // Verify that the requesting machine/user has at least trust level 3

    IVALIDATE_TRUST(3, "WAIT");
    
    // Parse the request
    
    STAFCommandParseResultPtr parsedResult =
        fWaitParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }                             

    DEFINE_VAR_POOL_LIST(varPoolList, varPoolListSize, requestInfo);
    STAFString errorBuffer;
    STAFString result;
    STAFString semName;
    STAFRC_t rc = RESOLVE_STRING_OPTION("EVENT", semName);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    STAFString uppercaseSemName = semName.toUpperCase();

    unsigned int timeout = STAF_EVENT_SEM_INDEFINITE_WAIT;
    
    rc = RESOLVE_DEFAULT_UINT_OPTION("TIMEOUT", timeout,
                                     STAF_EVENT_SEM_INDEFINITE_WAIT);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    EventUnit us = EventUnit(requestInfo.fMachine,
                             requestInfo.fHandleName,
                             requestInfo.fHandle,
                             requestInfo.fUser,
                             requestInfo.fEndpoint);
    EventSemPtr sem;

    fEventSemListSem.request();

    if (fEventSemList.find(uppercaseSemName) == fEventSemList.end())
    {
        // Semaphore does not exist, so create it in a reset state

        sem = EventSemPtr(new EventSem(semName), EventSemPtr::INIT);

        sem->lastReset = us;
        sem->eventSem->reset();

        fEventSemList[uppercaseSemName] = sem;
    }
    else
    {
        // The semaphore does exist, so get it.

        sem = fEventSemList[uppercaseSemName];
    }

    // Lock the semaphore, and unlock the main list.

    sem->lockPtr->request();
    fEventSemListSem.release();

    // First check to make sure that the semaphore is not already posted.
    // If it is posted, simply unlock the semaphore and return.

    if (sem->eventSem->query() == kSTAFEventSemPosted)
    {
        sem->lockPtr->release();
        return kSTAFOk;
    }

    // Ok, the semaphore is reset, so add us to the waiting list, and
    // unlock the semaphore.

    sem->waiterList.push_front(us);
    sem->lockPtr->release();

    // Now wait on it

    unsigned int waitRC = sem->eventSem->wait(timeout);

    // Re-acquire our lock on the semaphore

    sem->lockPtr->request();

    if (waitRC == 0)
    {
        // The sem was posted while we were trying to get a lock.
        // Yeah!!  We were removed from the waiting list by the
        // posting thread.

        sem->lockPtr->release();
    }
    else
    {
        // We timed out waiting for the semaphore, so remove
        // us from the waiting list

        int foundUs = 0;
        EventWaiterList::iterator iter;

        for(iter = sem->waiterList.begin(); !foundUs && 
            iter != sem->waiterList.end(); iter++)
        {
            if ((iter->machine == requestInfo.fMachine) &&
                (iter->handleName == requestInfo.fHandleName) &&
                (iter->handle == requestInfo.fHandle))
            {
                foundUs = 1;
                sem->waiterList.erase(iter);
                break;
            }
        }

        // Now unlock the semaphore and return a timeout error

        sem->lockPtr->release();

        return kSTAFTimeout;
    }

    return STAFServiceResult(kSTAFOk, result);
}

STAFServiceResult STAFSemService::handleDelete(
    const STAFServiceRequest &requestInfo)
{
    // Verify that the requesting machine/user has at least trust level 4

    IVALIDATE_TRUST(4, "DELETE");

    // Parse the request

    STAFCommandParseResultPtr parsedResult =
        fDeleteParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }

    DEFINE_VAR_POOL_LIST(varPoolList, varPoolListSize, requestInfo);
    STAFString errorBuffer;
    STAFString result;
    STAFString semName;

    if (parsedResult->optionTimes("MUTEX") != 0)
    {
        STAFRC_t rc = RESOLVE_STRING_OPTION("MUTEX", semName);

        if (rc) return STAFServiceResult(rc, errorBuffer);

        STAFString uppercaseSemName = semName.toUpperCase();

        STAFMutexSemLock lock(fMutexSemListSem);

        if (fMutexSemList.find(uppercaseSemName) == fMutexSemList.end())
            return kSTAFSemaphoreDoesNotExist;

        // Ok, the semaphore exists.  Get it and lock it.

        MutexSemPtr sem = fMutexSemList[uppercaseSemName];
        STAFMutexSemLock semLock(*sem->lockPtr);

        if (!sem->isOwned)
            fMutexSemList.erase(uppercaseSemName);
        else
            return kSTAFSemaphoreHasPendingRequests;
    }
    else
    {
        STAFRC_t rc = RESOLVE_STRING_OPTION("EVENT", semName);

        if (rc) return STAFServiceResult(rc, errorBuffer);

        STAFString uppercaseSemName = semName.toUpperCase();

        STAFMutexSemLock lock(fEventSemListSem);

        if (fEventSemList.find(uppercaseSemName) == fEventSemList.end())
            return kSTAFSemaphoreDoesNotExist;

        // Ok, the semaphore exists.  Get it and lock it.

        EventSemPtr sem = fEventSemList[uppercaseSemName];
        STAFMutexSemLock semLock(*sem->lockPtr);

        if (sem->waiterList.empty())
            fEventSemList.erase(uppercaseSemName);
        else
            return kSTAFSemaphoreHasPendingRequests;
    }
    
    return STAFServiceResult(kSTAFOk, result);
}

STAFServiceResult STAFSemService::handleQuery(
    const STAFServiceRequest &requestInfo)
{
    // Verify that the requesting machine/user has at least trust level 2

    IVALIDATE_TRUST(2, "QUERY");

    // Parse the request

    STAFCommandParseResultPtr parsedResult =
        fQueryParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }

    DEFINE_VAR_POOL_LIST(varPoolList, varPoolListSize, requestInfo);
    STAFString errorBuffer;
    STAFString semName;

    STAFObjectPtr mc = STAFObject::createMarshallingContext();

    if (parsedResult->optionTimes("MUTEX") != 0)
    {
        STAFRC_t rc = RESOLVE_STRING_OPTION("MUTEX", semName);

        if (rc) return STAFServiceResult(rc, errorBuffer);

        STAFString uppercaseSemName = semName.toUpperCase();
        
        fMutexSemListSem.request();

        if (fMutexSemList.find(uppercaseSemName) == fMutexSemList.end())
        {
            fMutexSemListSem.release();
            return kSTAFSemaphoreDoesNotExist;
        }

        // Create a marshalled map containing information about the semaphore

        mc->setMapClassDefinition(fQueryMutexClass->reference());
        mc->setMapClassDefinition(fMutexOwnerClass->reference());
        mc->setMapClassDefinition(fPendingRequestClass->reference());
        STAFObjectPtr mutexMap = fQueryMutexClass->createInstance();

        // Ok, the semaphore exists.  Get it and lock it, then release the
        // lock on the semaphore list.

        MutexSemPtr sem = fMutexSemList[uppercaseSemName];
        STAFMutexSemLock semLock(*sem->lockPtr);

        fMutexSemListSem.release();

        STAFObjectPtr pendingRequests = STAFObject::createList();

        if (sem->isOwned)
        {
            mutexMap->put("state", "Owned");

            STAFObjectPtr ownerMap = fMutexOwnerClass->createInstance();
            ownerMap->put("machine", sem->owner.machine);
            ownerMap->put("handleName", sem->owner.handleName);
            ownerMap->put("handle", STAFString(sem->owner.handle));
            ownerMap->put("user", sem->owner.user);
            ownerMap->put("endpoint", sem->owner.endpoint);
            ownerMap->put("requestTimestamp",
                          sem->owner.timeDate.asString());
            ownerMap->put("acquireTimestamp",
                          sem->acquireTimeDate.asString());

            mutexMap->put("owner", ownerMap);

            if (!sem->requestList.empty())
            {
                MutexRequestList::iterator iter;
                int i = 1;

                for(iter = sem->requestList.begin();
                    iter != sem->requestList.end();
                    ++i, ++iter)
                {
                    MutexRequest request = *iter;

                    STAFObjectPtr requestMap =
                        fPendingRequestClass->createInstance();

                    requestMap->put("machine", request.machine);
                    requestMap->put("handleName", request.handleName);
                    requestMap->put("handle", STAFString(request.handle));
                    requestMap->put("user", request.user);
                    requestMap->put("endpoint", request.endpoint);
                    requestMap->put("requestTimestamp",
                                    request.timeDate.asString());

                    pendingRequests->append(requestMap);
                }
            }

        }
        else
        {
            mutexMap->put("state", "Unowned");
        }

        mutexMap->put("requestList", pendingRequests);

        mc->setRootObject(mutexMap);
    }
    else
    {
        STAFRC_t rc = RESOLVE_STRING_OPTION("EVENT", semName);

        if (rc) return STAFServiceResult(rc, errorBuffer);

        STAFString uppercaseSemName = semName.toUpperCase();

        fEventSemListSem.request();

        if (fEventSemList.find(uppercaseSemName) == fEventSemList.end())
        {
            fEventSemListSem.release();
            return kSTAFSemaphoreDoesNotExist;
        }

        // Create a marshalled map containing information about the semaphore

        mc->setMapClassDefinition(fQueryEventClass->reference());
        mc->setMapClassDefinition(fEventRequesterClass->reference());
        STAFObjectPtr eventMap = fQueryEventClass->createInstance();

        // Ok, the semaphore exists.  Get it and lock it, then release the
        // lock on the semaphore list.

        EventSemPtr sem = fEventSemList[uppercaseSemName];
        STAFMutexSemLock lock(*sem->lockPtr);

        fEventSemListSem.release();

        if (sem->eventSem->query() == kSTAFEventSemPosted)
            eventMap->put("state", STAFString("Posted"));
        else
            eventMap->put("state", STAFString("Reset"));

        STAFObjectPtr lastPostedMap = fEventRequesterClass->createInstance();

        lastPostedMap->put("machine", sem->lastPost.machine);
        lastPostedMap->put("handleName", sem->lastPost.handleName);
        lastPostedMap->put("handle", STAFString(sem->lastPost.handle));
        lastPostedMap->put("user", sem->lastPost.user);
        lastPostedMap->put("endpoint", sem->lastPost.endpoint);
        lastPostedMap->put("timestamp", sem->lastPost.timeDate.asString());

        eventMap->put("lastPosted", lastPostedMap);

        STAFObjectPtr lastResetMap = fEventRequesterClass->createInstance();

        lastResetMap->put("machine", sem->lastReset.machine);
        lastResetMap->put("handleName", sem->lastReset.handleName);
        lastResetMap->put("handle", STAFString(sem->lastReset.handle));
        lastResetMap->put("user", sem->lastReset.user);
        lastResetMap->put("endpoint", sem->lastReset.endpoint);
        lastResetMap->put("timestamp", sem->lastReset.timeDate.asString());

        eventMap->put("lastReset", lastResetMap);
        
        STAFObjectPtr outputList = STAFObject::createList();

        if (sem->eventSem->query() == kSTAFEventSemReset)
        {
            if (sem->waiterList.size() != 0)
            {
                EventWaiterList::iterator iter;
                int i = 1;

                for(iter = sem->waiterList.begin();
                    iter != sem->waiterList.end(); ++i, ++iter)
                {
                    EventUnit eventUnit = *iter;

                    STAFObjectPtr waiterMap =
                        fEventRequesterClass->createInstance();
                    waiterMap->put("machine", eventUnit.machine);
                    waiterMap->put("handleName", eventUnit.handleName);
                    waiterMap->put("handle", STAFString(eventUnit.handle));
                    waiterMap->put("user", eventUnit.user);
                    waiterMap->put("endpoint", eventUnit.endpoint);
                    waiterMap->put("timestamp", eventUnit.timeDate.asString());

                    outputList->append(waiterMap);
                }
            }

        }

        eventMap->put("waiterList", outputList);

        mc->setRootObject(eventMap);
    }

    return STAFServiceResult(kSTAFOk, mc->marshall());
}

STAFServiceResult STAFSemService::handleList(
    const STAFServiceRequest &requestInfo)
{
    // Verify that the requesting machine/user has at least trust level 2

    IVALIDATE_TRUST(2, "LIST");

    // Parse the request

    STAFCommandParseResultPtr parsedResult = fListParser.parse(
        requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }

    STAFObjectPtr mc = STAFObject::createMarshallingContext();

    if (parsedResult->optionTimes("MUTEX") != 0)
    {
        // Create a marshalled list of maps containing mutex semaphore info

        mc->setMapClassDefinition(fMutexInfoClass->reference());
        STAFObjectPtr outputList = STAFObject::createList();
        
        STAFMutexSemLock lock(fMutexSemListSem);
        MutexSemList::iterator iter;

        for(iter = fMutexSemList.begin(); iter != fMutexSemList.end(); iter++)
        {
            MutexSemPtr sem = iter->second;
            STAFMutexSemLock semLock(*sem->lockPtr);

            STAFObjectPtr mutexInfoMap = fMutexInfoClass->createInstance();

            mutexInfoMap->put("name", STAFString(sem->name));

            if (sem->isOwned)
            {
                mutexInfoMap->put("state", STAFString("Owned"));
            }
            else
            {
                mutexInfoMap->put("state", STAFString("Unowned"));
            }

            mutexInfoMap->put("pendingRequests",
                              STAFString(sem->requestList.size()));

            outputList->append(mutexInfoMap);
        }

        mc->setRootObject(outputList);
    }
    else
    {
       // Create a marshalled list of maps containing event semaphore info

        mc->setMapClassDefinition(fEventInfoClass->reference());
        STAFObjectPtr outputList = STAFObject::createList();

        STAFMutexSemLock lock(fEventSemListSem);
        EventSemList::iterator iter;
        
        for(iter = fEventSemList.begin(); iter != fEventSemList.end(); iter++)
        {
            EventSemPtr sem = iter->second;
            STAFMutexSemLock semLock(*sem->lockPtr);

            STAFObjectPtr eventInfoMap = fEventInfoClass->createInstance();

            eventInfoMap->put("name", STAFString(sem->name));

            if (sem->eventSem->query() == kSTAFEventSemReset)
            {
                eventInfoMap->put("state", STAFString("Reset"));
            }
            else
            {
                eventInfoMap->put("state", STAFString("Posted"));
            }

            eventInfoMap->put("waiters", STAFString(sem->waiterList.size()));

            outputList->append(eventInfoMap);
        }

        mc->setRootObject(outputList);
    }

    return STAFServiceResult(kSTAFOk, mc->marshall());
}


STAFServiceResult STAFSemService::handleSTAFCallback(
    const STAFServiceRequest &requestInfo)
{
    STAFString result = " ";

    // Don't check the TRUST level, but make sure the requesting handle
    // is the STAFProc handle

    if (requestInfo.fHandle != 1)
    {
        // XXX Return a non-zero RC?
        return STAFServiceResult(kSTAFOk, result);
    }
    
    if (debug) // XXX
        cout << "handleSTAFCallback in SEM service\n";
    
    STAFCommandParseResultPtr parsedResult =
        fSTAFCallbackParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }
    
    STAFString handle = parsedResult->optionValue("HANDLE");
    STAFString machine = parsedResult->optionValue("MACHINE");
    STAFString uuid = parsedResult->optionValue("UUID");
    
    STAFMutexSemLock lock(fMutexSemListSem);
    MutexSemList::iterator iter;

    for(iter = fMutexSemList.begin(); iter != fMutexSemList.end(); iter++)
    {
        MutexSemPtr sem = iter->second;

        if (debug) // XXX
            cout << sem->name << " owned by " << sem->owner.handle
                 << " callbackHandle: " << handle << "\n";
                 
        if (debug) // XXX
            cout << sem->name << " owned by " << sem->owner.machine
                 << " callbackMachine: " << machine << "\n";
        
        if ((STAFString(sem->owner.handle) == handle) && 
             (STAFString(sem->owner.machine) == machine) && 
             (STAFString(sem->owner.stafInstanceUUID) == uuid))
        {
            if (debug) // XXX
                cout << "callback found a matching handle: " << handle << "\n";
         
            MutexSemPtr semaphore = fMutexSemList[sem->uppercaseName];
            STAFMutexSemLock semLock(*semaphore->lockPtr);

            // Verify we are allowed to release the semaphore
            
            if (semaphore->requestList.size() != 0)
            {
                // There is a pending request.  Remove it from the list and
                // notify the waiter.
    
                semaphore->acquireTimeDate = STAFTimestamp::now();
    
                semaphore->owner = semaphore->requestList.front();
                semaphore->requestList.pop_front();

                semaphore->owner.avail->post();
            }
            else 
            {
                semaphore->isOwned = 0;
            }
        }
    }

    return STAFServiceResult(kSTAFOk, result);
}


STAFServiceResult STAFSemService::handleHelp(
    const STAFServiceRequest &requestInfo)
{
    // Verify that the requesting machine/user has at least trust level 1

    IVALIDATE_TRUST(1, "HELP");

    STAFString result("SEM (Semaphore) Service Help" + *gLineSeparatorPtr +
                       *gLineSeparatorPtr);

    result += "REQUEST MUTEX <Name> [TIMEOUT <Timeout>]" +
              *gLineSeparatorPtr;
    result += "RELEASE MUTEX <Name> [FORCE]" + *gLineSeparatorPtr +
              *gLineSeparatorPtr;
    result += "POST    EVENT <Name>" + *gLineSeparatorPtr;
    result += "RESET   EVENT <Name>" + *gLineSeparatorPtr;
    result += "PULSE   EVENT <Name>" + *gLineSeparatorPtr;
    result += "WAIT    EVENT <Name> [TIMEOUT <Timeout>]" + *gLineSeparatorPtr +
              *gLineSeparatorPtr;
    result += "DELETE  MUTEX <Name> | Event <Name>" + *gLineSeparatorPtr;
    result += "QUERY   MUTEX <Name | Event <Name>" + *gLineSeparatorPtr;
    result += "LIST    MUTEX | EVENT" + *gLineSeparatorPtr;
    result += "HELP" + *gLineSeparatorPtr;

    return STAFServiceResult(kSTAFOk, result);
}
